# This class provides a very simple interface for a hash with multiple keys
# that reference the same value. This is convenient for lookup tables that have
# several mappings to a single value, without doing anything particularly
# gross with Array#include? lookups.
class MultiHash
  attr_accessor :immutable
  alias_method :immutable?, :immutable

  HASH_METHOD_WHITELIST = %w{[] assoc default default= empty? each each_key
    each_pair each_value fetch keys select values_at}

  HASH_METHOD_WHITELIST.each do |mth|
    define_method mth do |*args, &blk|
      @_hash.send(mth, *args, &blk)
    end
  end

  # Initialize an immutable multihash with the provided hash as its source.
  # For example, we might want to look up the nation which produces different
  # types of cars:
  #
  #     MultiHash.new (
  #       %w{Honda Nissan} => :japan,
  #       %w{Ford Chrysler} => :us,
  #       'BMW' => :germany
  #     )
  #
  # @param original_hash [Hash<Array|Object, Object>] the hash to convert into
  # a multihash, with arrays or single objects as keys.
  def initialize(original_hash={}, immutable=true)
    @_hash = Hash.new(original_hash.default)
    @immutable = immutable

    groups = []

    original_hash.each do |orig_key, value|
      key_array = orig_key.is_a?(Array) && orig_key || [orig_key]
      groups << key_array
      key_array.each { |k| map_value(k, value) }
    end

    store_groups(groups) unless immutable?
  end

  def grouped
    Hash[@_groups.values.map { |g| [g, self[g.first]] }]
  end

  def set(key, value)
    fail 'Cannot modify an immutable MultiHash!' if immutable?
    split_key_from_group(key)
    map_value(key, value)
  end
  alias_method :[]=, :set

  def set_group(key, value)
    fail 'Cannot modify an immutable MultiHash!' if immutable?
    @_groups[key].each { |k| map_value(k, value) }
  end

  def to_h
    @_hash
  end

  def values
    @_values ||= @_hash.values.uniq
  end

  def self.[](args)
    MultiHash.new(Hash[args])
  end

  private

  def dirty
    @_grouped = nil
    @_values = nil
  end

  def map_value(key, value)
    dirty
    @_hash[key] = value
  end

  def split_key_from_group(key)
    former_group = @_groups[key]
    other_groups = @_groups.values - former_group
    new_group = key.is_a?(Array) ? key : [key]
    modified_group = former_group - new_group
    store_groups(other_groups + modified_group + new_group)
  end

  def store_groups(groups)
    @_groups = MultiHash.new(Hash[groups.map { |g| [g, g] }])
  end
end
